%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% gprolog source code for Interactive Klumpenhouwer Networks
%% by Israel Neuman and Alberto M. Segre
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% need to be compile to object code using gplc -c and then linking
%% it into PD.

%% To use negation as failure, must set prolog directive to ensure
%% unknown clauses fail and don't instead throw an error.
:- set_prolog_flag(unknown,fail).

:- dynamic(hyperI/4).
:- dynamic(hyperT/4).
:- dynamic(aKNet/2).

%% GProlog gensym predicate. 
%% From: http://www.david-reitter.com/compling/prolog/compat_gnu.pl
flag(Sym, Old, NewExp) :-
	g_read(Sym, Old),
	nonvar(NewExp) ->
		New is NewExp,  % evaluate
		g_assign(Sym, New)
		;
		g_read(Sym, New).
		
% Invoked as gensym(a,P) will return a new atom
% prefixed with a each time.
gensym(Base, Atom) :-
	atom_concat('gs_', Base, Key), 
	flag(Key, N, N+1),
	number_atom(N, NA),
	atom_concat(Base, NA, Atom).

%% Negation as failure.
neg(G) :- G, !, fail.
neg(_).

%% Count occurrences of X in Z.
count(_,[],0).
count(X,[X|Z],N) :- count(X,Z,N2), N is N2 + 1.
count(X,[Y|Z],N) :- X\==Y, count(X,Z,N). 

%% N<=M: both N and M must be bound.
limit(N,M) :- N>=M,!,fail.
limit(_,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Analysis for isographs
trnPitch(P,N1,X) :- X is mod((P+N1),12).
invPitch(P,N2,Y) :- Y is mod((-P+N2),12).

invertA([X,Y|_],InvA) :- InvA is mod((X+Y),12). 
invertB([_,Y|Z],InvB) :- InvB is mod((Y+Z),12).
invertBB([_,Y,Z|_],InvBB) :- InvBB is mod((Y+Z),12).
invertC([_,_,Y|Z],InvC) :- InvC is mod((Y+Z),12).
transposeA([X,_|Z],TrnA) :- TrnA is mod((Z-X),12).
transposeAA([W,_,_|Z],TrnAA) :- TrnAA is mod((Z-W),12).
transposeB([_,_,X|Y],TrnB) :- TrnB is mod((Y-X),12).
transposeBB([_,X,Y|_],TrnBB) :- TrnBB is mod((Y-X),12).

analyzeTI3(Set,[InvA,InvB,TrnA]) :- invertA(Set,InvA), invertB(Set,InvB), transposeA(Set,TrnA).
analyzeTI4(Set,[InvA,TrnBB,InvC,TrnAA]) :- invertA(Set,InvA), transposeBB(Set,TrnBB), 
	invertC(Set,InvC), transposeAA(Set,TrnAA).

genIsoG([X,Y,Z],U,J,P,[P1,P2,P3]) :- analyzeTI3([X,Y,Z],[InvA,InvB,_]),P1 is P, 
	invPitch(P1,(InvA*U+J),P2),invPitch(P2,(InvB*U+J),P3). 

genIsoG([W,X,Y,Z],U,J,P,[P1,P2,P3,P4]) :- analyzeTI4([W,X,Y,Z],[InvA,TrnBB,InvC,_]), 
	P1 is P, invPitch(P1,(InvA*U+J),P2),
	trnPitch(P2,(TrnBB*U),P3), invPitch(P3,(InvC*U+J),P4). 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% In first phase, user will issue one or more genAndAssert()
%% statements. This is the top level call to generate a network.

%% Creates an aKNet with a unique label L, and also generate its associated hyper statements
genAndAssert(Set, U, J, P, S) :- genKNet(Set, U, J, P, S), neg(aKNet([Set | S], _)), gensym(aknet,L),
	assertz(aKNet([Set | S], L)), genHyperStmts(Set, 1, 0, S, U, J, L).

genKNet(_,[],[],[],[]).
genKNet(Set,[UH|UT],[JH|JT],[PH|PT],[SH|ST]) :- genIsoG(Set,UH,JH,PH,SH), genKNet(Set,UT,JT,PT,ST).

%% genAndAssert() inserts aKNet() statements in the DB, which then can
%% be used to derive hyperI() and hyperT() statements. The aKNet()
%% statements are really only there to make it easier for composer to
%% see what the internal structure of the networks
%% are. genHyperStmts() takes care of adding all the appropriate
%% hyperI() and hyperT() statments. Note that each aKNet is labeled
%% with a unique label, which is also used in each hyperI() and hyperT()
%% statement. We can use the label to make sure we only use hyperI and hyperT
%% statements from the same aKNet together.
genHyperStmts(S, US, JS, [SH | ST], [UH | UT], [JH | JT], L) :- addHyperStmt(S, SH, US, UH, JS, JH, L), 
	genHyperStmts(S, US, JS, ST, UT, JT, L), genHyperStmts(SH, UH, JH, ST, UT, JT, L).
genHyperStmts(_, _, _, [], [], [], _).

%% Compute the type of hyper statement and the appropriate
%% values. Note use of negation as failure again to ensure no
%% duplicate hyper statements get added to the database.
addHyperStmt(S1, S2, U1, U2, J1, J2, L) :- U1 \= U2, F1 is mod(((U1*J2)+J1),12), neg(hyperI(S1, S2, F1, L)), assertz(hyperI(S1, S2, F1, L)).
addHyperStmt(S1, S2, U, U, J1, J2, L) :- F2 is mod(((U*J2)+J1),12), neg(hyperT(S1, S2, F2, L)), assertz(hyperT(S1, S2, F2, L)).
%% Because we are using neg() to check for duplicates, and because
%% addHyperStmt() must always succeed (so that genHyperStmts() succeeds) in
%% order to ensure we explore the entire space of the graph, we need an
%% extra clause that allows addHyperStmt() to succeed even if it was a
%% duplicate call.
addHyperStmt(_, _, _, _, _, _, _).

%% Fake hyperI and hyperT to satisfy the C compiler.
%% hyperI([],[],101,foobar).
%% hyperT([],[],102,foobaz).

%% Some test functions for phase one (remove later)
testa(L) :- genAndAssert([3,10,4],[11,11,11,11,1],[2,1,8,2,6],[4,10,5,10,6],L).
testb :- genAndAssert([11,3,10],[1,1,1],[0,9,0],[2,7,9],_).
testCookA :- genAndAssert([8,5,2],[11],[8],[2],_).
testCookB :- genAndAssert([1,9,5],[11],[0],[5],_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% In the second phase, user will issue one or more chain()
%% statements. This is the top level call to generate a Path from a
%% known Graph with optional Start and Destination sets. Use
%% a random Seed value to alter choice points in the search tree.
findPath(Src,Dst,Path,Graph,Horizon,Repeats) :- random(0,2,Seed),modchain(Src,Dst,Path,Graph,1,Seed,Horizon,Repeats).

%% These rules define a tree, rooted in hyperI/hyperT statements. We
%% are looking for a random path in this tree, meaning that we need to
%% pick a random branch at each choice point as well as a random
%% instantiations from among the legal instantiations at the root
%% hyperI/hyperT statements.

%% The second is simple: use a (randomly seeded) random number
%% generator to elect which of three legal paths to take at each call
%% to related(). We use modrelated() to step through all three related
%% clauses in some ordered, determined by the input random "seed", S.
modrelated(X,Y,T,L,S) :- N is mod(S,3),related(X,Y,T,L,N).
modrelated(X,Y,T,L,S) :- N is mod(S+1,3),related(X,Y,T,L,N).
modrelated(X,Y,T,L,S) :- N is mod(S+2,3),related(X,Y,T,L,N).
%% Base related clauses, meant to be tried in some random order.
related(X,Y,T,L,0) :- hyperT(X,Y,T,L).
related(X,Y,I,L,1) :- hyperI(X,Y,I,L).
related(X,Y,I,L,2) :- hyperI(Y,X,I,L).

%% The chain rules are a bit harder. If we impose a fixed horizon
%% (say, H=4, which means 4 recursive steps and a total chain length
%% of 5 or less), we can try the two chain rules, one recursive and
%% one not, in random order, as we know the search remains finite. We
%% should also probably fail on too many replicates in a chain by
%% checking X against Z before adding X to Z in the recursive step,
%% using the magic number 3 as the upper limit on replicates in a
%% chain.
modchain(X,Y,W,L,D,S,H,R) :- N is mod(S,2),chain(X,Y,W,L,D,N,H,R).
modchain(X,Y,W,L,D,S,H,R) :- N is mod(S+1,2),chain(X,Y,W,L,D,N,H,R).
%% Base chain clause.
chain(_,_,_,_,D,_,H,_) :- D>H,!,fail.
chain(X,Y,[X,Y],L,_,0,_,_) :- random(0,3,S),modrelated(X,Y,_,L,S).
chain(X,Y,[X|Z],L,D,1,H,R) :- random(0,3,S),modrelated(X,W,_,L,S),D2 is D+1,random(0,2,S2),modchain(W,Y,Z,L,D2,S2,H,R),count(X,Z,C),limit(C,R).

%% Here's an example: 
%%   To find a path from <src> to <dst> of length no more 
%%   than 4 steps and with no more than 2 copies of any triple:
%% findPath(Src,Dst,Path,aknet0,4,2).
